from django.test import TestCase

from buybackprogram.helpers import get_tracking_number

from .testdata.factories import ProgramFactory, UserIssuerFactory
from .testdata.load_eveuniverse import load_eveuniverse


class TestHelpers(TestCase):
    def test_should_create_tracking_when_first(self):
        # given
        load_eveuniverse()
        user = UserIssuerFactory()
        program = ProgramFactory()
        contract_net_prices = {
            "total_all_items_raw": 1,
            "total_all_items": 1,
            "total_tax_amount": 1,
            "total_donation_amount": 1,
            "hauling_cost": 0,
            "total_hauling_cost": 0,
            "contract_net_total": 1,
        }
        # when
        tracking = get_tracking_number(user, program, None, [], contract_net_prices)
        # then
        self.assertEqual(tracking.program, program)
        self.assertEqual(tracking.issuer_user, user)



from datetime import timedelta

import requests
from bravado.exception import HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
from celery import shared_task

from django.db import Error
from django.utils import timezone
from eveuniverse.models import EveMarketPrice

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

from buybackprogram.app_settings import (
    BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
    BUYBACKPROGRAM_PRICE_METHOD,
    BUYBACKPROGRAM_PRICE_SOURCE_ID,
    BUYBACKPROGRAM_PRICE_SOURCE_NAME,
)
from buybackprogram.models import ItemPrices, Owner, Tracking

from .app_settings import (
    BUYBACKPROGRAM_TASKS_TIME_LIMIT,
    BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT,
)

logger = get_extension_logger(__name__)

NORMAL_TASK_PRIORITY = 4

# Create your tasks here
TASK_DEFAULT_KWARGS = {
    "time_limit": BUYBACKPROGRAM_TASKS_TIME_LIMIT,
}

TASK_ESI_KWARGS = {
    **TASK_DEFAULT_KWARGS,
    **{
        "bind": True,
        "autoretry_for": (
            OSError,
            HTTPBadGateway,
            HTTPGatewayTimeout,
            HTTPServiceUnavailable,
        ),
        "retry_kwargs": {"max_retries": 3},
        "retry_backoff": 30,
    },
}


def valid_janice_api_key():
    c = requests.get(
        "https://janice.e-351.com/api/rest/v2/markets",
        headers={
            "Content-Type": "text/plain",
            "X-ApiKey": BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
            "accept": "application/json",
        },
    ).json()

    if "status" in c:
        logger.debug("Janice API status: %s" % c)
        return False
    else:
        return True


def get_bulk_prices(type_ids):
    r = None
    if BUYBACKPROGRAM_PRICE_METHOD == "Fuzzwork":
        r = requests.get(
            "https://market.fuzzwork.co.uk/aggregates/",
            params={
                "types": ",".join([str(x) for x in type_ids]),
                "station": BUYBACKPROGRAM_PRICE_SOURCE_ID,
            },
        ).json()
    elif BUYBACKPROGRAM_PRICE_METHOD == "Janice":
        r = requests.post(
            "https://janice.e-351.com/api/rest/v2/pricer?market=2",
            data="\n".join([str(x) for x in type_ids]),
            headers={
                "Content-Type": "text/plain",
                "X-ApiKey": BUYBACKPROGRAM_PRICE_JANICE_API_KEY,
                "accept": "application/json",
            },
        ).json()

        # Make Janice data look like Fuzzworks
        output = {}
        for item in r:
            output[str(item["itemType"]["eid"])] = {
                "buy": {"max": str(item["top5AveragePrices"]["buyPrice5DayMedian"])},
                "sell": {"min": str(item["top5AveragePrices"]["sellPrice5DayMedian"])},
            }
        r = output
    else:
        raise f"Unknown pricing method: {BUYBACKPROGRAM_PRICE_METHOD}"

    return r


@shared_task
def update_all_prices():
    type_ids = []
    market_data = {}
    api_up = True

    # Get all type ids
    prices = ItemPrices.objects.all()

    if BUYBACKPROGRAM_PRICE_METHOD == "Fuzzwork":
        logger.debug(
            "Price setup starting for %s items from Fuzzworks API from station id %s (%s), this may take up to 30 seconds..."
            % (
                len(prices),
                BUYBACKPROGRAM_PRICE_SOURCE_ID,
                BUYBACKPROGRAM_PRICE_SOURCE_NAME,
            )
        )
    elif BUYBACKPROGRAM_PRICE_METHOD == "Janice":
        if valid_janice_api_key():
            logger.debug(
                "Price setup starting for %s items from Janice API for Jita 4-4, this may take up to 30 seconds..."
                % (len(prices),)
            )
        else:
            logger.debug(
                "Price setup failed for Janice, invalid API key! Provide a working key or change price source to Fuzzwork"
            )
            api_up = False
    else:
        logger.error(
            "Unknown pricing method: '%s', skipping" % BUYBACKPROGRAM_PRICE_METHOD
        )
        return

    if api_up:
        # Build suitable bulks to fetch prices from API
        for item in prices:
            type_ids.append(item.eve_type_id)

            if len(type_ids) == 1000:
                market_data.update(get_bulk_prices(type_ids))
                type_ids.clear()

        # Get leftover data from the bulk
        market_data.update(get_bulk_prices(type_ids))

        logger.debug("Market data fetched, starting database update...")
        missing_items = []
        for price in prices:
            # Check if we received data from the API for the item. This will fix errors when using Janice as the POST endpoint does not return anything when there are no prices.
            if str(price.eve_type_id) in market_data:
                # Get the price values from the API data
                buy = int(float(market_data[str(price.eve_type_id)]["buy"]["max"]))
                sell = int(float(market_data[str(price.eve_type_id)]["sell"]["min"]))

            # If API did not return any values we remove prices for the item
            else:
                missing_items.append(price.eve_type.name)

                # Reset prices for items not found from API
                buy = 0
                sell = 0

            price.buy = buy
            price.sell = sell
            price.updated = timezone.now()

        try:
            ItemPrices.objects.bulk_update(prices, ["buy", "sell", "updated"])
            logger.debug("All prices succesfully updated")
        except Error as e:
            logger.error("Error updating prices: %s" % e)

        EveMarketPrice.objects.update_from_esi()

        logger.debug("Updated all eveuniverse market prices.")

        if len(missing_items) > 0:
            logger.error(
                "%s items missing items from source API, prices set to 0."
                % len(missing_items)
            )

    else:
        logger.error("Price source API is not up! Prices not updated.")

    if BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT > 0:
        """cleanup unused tracking objects"""
        logger.debug(
            "Starting tracking objects cleanup. Removing tracking objects with no contracts assigned to them that are more than %s hours old "
            % BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT
        )

        try:
            trackings, t = Tracking.objects.filter(
                contract_id__isnull=True,
                created_at__lte=timezone.now()
                - timedelta(hours=BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT),
            ).delete()
            logger.debug("%s old unlinked tracking objects deleted." % len(t))
        except Error as e:
            logger.error("Error deleting old tracking objects: %s" % e)
    else:
        logger.debug(
            "Tracking object time limit is set to %s, no old tracking object purging will happen. "
            % BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT
        )


@shared_task(
    **{
        **TASK_ESI_KWARGS,
        **{
            "base": QueueOnce,
            "once": {"keys": ["owner_pk"], "graceful": True},
            "max_retries": None,
        },
    }
)
def update_contracts_for_owner(self, owner_pk):
    """fetches all contracts for owner from ESI"""

    return _get_owner(owner_pk).update_contracts_esi()


@shared_task(**TASK_DEFAULT_KWARGS)
def update_all_contracts():
    logger.debug("Starting all contract updates")
    for owner in Owner.objects.all():
        logger.debug("Updating contracts for %s" % owner)
        update_contracts_for_owner.apply_async(
            kwargs={"owner_pk": owner.pk},
            priority=NORMAL_TASK_PRIORITY,
        )


def _get_owner(owner_pk: int) -> Owner:
    """returns the owner or raises exception"""
    try:
        owner = Owner.objects.get(pk=owner_pk)
    except Owner.DoesNotExist:
        raise Owner.DoesNotExist(
            "Requested owner with pk {} does not exist".format(owner_pk)
        )
    return owner



# Generated by Django 3.2.10 on 2022-01-20 12:25

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="ContractNotification",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("icon", models.CharField(max_length=64)),
                ("color", models.CharField(max_length=32)),
                ("message", models.CharField(max_length=1024)),
                (
                    "contract",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.contract",
                    ),
                ),
            ],
        ),
    ]



from django.core.management.base import BaseCommand
from django.db import Error

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.models import Contract, Tracking

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = "Links tracking objects with old contracts after 1.2.0 update"

    def handle(self, *args, **options):
        updated = 0

        self.stdout.write("Starting contract and tracking manual linking...")

        # Get all contracts for owner
        all_contracts = Contract.objects.all()

        self.stdout.write(
            "Found %s old contracts stored in database" % len(all_contracts)
        )

        # Get all tracking objects from the database
        tracking_numbers = Tracking.objects.filter(contract__isnull=True)

        self.stdout.write(
            "Found %s tracking objects with no contracts" % len(tracking_numbers)
        )

        # Start looping for all stored tracking objects
        for tracking in tracking_numbers:
            # If the tracking has an active program (not deleted)
            if tracking.program:
                # Start checking if we find any matches from our ESI contracts
                for contract in all_contracts:
                    # Only get contracts with the correct prefill ticker
                    if tracking.tracking_number in contract.title:
                        try:
                            tracking = Tracking.objects.filter(pk=tracking.id).update(
                                contract=contract.id
                            )

                            updated += 1

                        except Error as e:
                            logger.error(
                                "Error linking contract %s with tracking %s: %s"
                                % (
                                    contract.contract_id,
                                    tracking.tracking_number,
                                    e,
                                )
                            )

                        break

        self.stdout.write("Linked up %s tracking objects with contracts" % (updated))



"""
notifications helper
"""
import requests
from django.contrib.humanize.templatetags.humanize import intcomma
from django.contrib.auth.models import User

from allianceauth.notifications import notify
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.app_settings import (
    aa_discordnotify_active,
    allianceauth_discordbot_active,
)

logger = get_extension_logger(__name__)


def send_aa_discordbot_notification(user, message):
    # If discordproxy app is not active we will check if aa-discordbot is active
    if allianceauth_discordbot_active():
        from aadiscordbot.tasks import send_message
        from discord import Embed

        contract = message["contract"]
        tracking = message["tracking"]

        embed = Embed(
            description=message["contract_items"],
            title=message["title"],
            color=message["color"],
        )

        embed.add_field(
            name="Date issued", value=str(contract.date_issued), inline=True
        )
        embed.add_field(
            name="Issued from", value=str(message["assigned_from"]), inline=True
        )
        embed.add_field(
            name="Issued to", value=str(message["assigned_to"]), inline=True
        )
        embed.add_field(name="Location", value=str(contract.location_name), inline=True)
        embed.add_field(
            name="Tracking #", value=str(tracking.tracking_number), inline=True
        )
        embed.add_field(
            name="Volume",
            value=str(intcomma(int(contract.volume))) + " m3",
            inline=True,
        )
        embed.add_field(
            name="Value", value=str(intcomma(int(contract.price))) + " ISK", inline=True
        )
        embed.add_field(name="Notes", value=str(message["notes"]), inline=False)

        send_message(user_pk=user, embed=embed)

        logger.debug("Sent discord DM to user %s" % user)
    else:
        logger.debug(
            "No discord notification modules active. Will not send user notifications"
        )


def send_aa_discordbot_channel_notification(channel_id, message):
    # If discordproxy app is not active we will check if aa-discordbot is active
    if allianceauth_discordbot_active():
        import aadiscordbot.tasks

        aadiscordbot.tasks.send_channel_message_by_discord_id.delay(
            channel_id, message, embed=True
        )

        logger.debug("Sent notification to channel %s" % channel_id)
    else:
        logger.debug(
            "No discord notification modules active. Will not send user notifications"
        )


# This will set a auth notification and send the user a direct notification on discord if requested for new contracts
def send_user_notification(user: User, level: str, message: dict) -> None:
    contract = message["contract"]
    tracking = message["tracking"]

    msg = (
        "Date issued: "
        + str(contract.date_issued)
        + "\n"
        + "Issued from: "
        + str(message["assigned_from"])
        + "\n"
        + "Location: "
        + str(contract.location_name)
        + "\n"
        + "Tracking #: "
        + str(tracking.tracking_number)
        + "\n"
        + "Price #: "
        + str(intcomma(int(contract.price)))
        + " ISK"
        + "\n"
    )
    notify(
        user=user,
        title=message["title"],
        level=level,
        message=msg,
    )

    if not aa_discordnotify_active():
        # Check if the discordproxy module is active. We will use it as our priority app for notifications
        try:
            from discordproxy.client import DiscordClient
            from discordproxy.discord_api_pb2 import Embed
            from discordproxy.exceptions import (
                DiscordProxyException,
                DiscordProxyGrpcError,
            )

            logger.debug("User has a active discord account")

            client = DiscordClient()

            contract = message["contract"]
            tracking = message["tracking"]

            fields = []

            fields.append(
                Embed.Field(
                    name="Date issued", value=str(contract.date_issued), inline=True
                )
            )
            fields.append(
                Embed.Field(
                    name="Issued from", value=str(message["assigned_from"]), inline=True
                )
            )
            fields.append(
                Embed.Field(
                    name="Issued to", value=str(message["assigned_to"]), inline=True
                )
            )
            fields.append(
                Embed.Field(
                    name="Location", value=str(contract.location_name), inline=True
                )
            )
            fields.append(
                Embed.Field(
                    name="Tracking #", value=str(tracking.tracking_number), inline=True
                )
            )
            fields.append(
                Embed.Field(
                    name="Volume",
                    value=str(intcomma(int(contract.volume))) + " m3",
                    inline=True,
                )
            )
            fields.append(
                Embed.Field(
                    name="Value",
                    value=str(intcomma(int(contract.price))) + " ISK",
                    inline=True,
                )
            )
            fields.append(
                Embed.Field(name="Notes", value=str(message["notes"]), inline=False)
            )

            embed = Embed(
                title=message["title"],
                color=message["color"],
                fields=fields,
                author=Embed.Author(name="AA Buyback Program"),
            )

            if user.discord.uid:
                try:
                    logger.debug(
                        "Sending notification for discord user %s" % user.discord.uid
                    )
                    client.create_direct_message(user_id=user.discord.uid, embed=embed)
                except DiscordProxyGrpcError:
                    logger.debug(
                        "Discordprox is installed but not running, failed to send message. Attempting to send via aa-discordbot instead."
                    )
                    send_aa_discordbot_notification(user.pk, message)

                except DiscordProxyException as ex:
                    logger.error(
                        "An error occured when trying to create a message: %s" % ex
                    )
            else:
                logger.error(
                    "User has no active discord accounts on AUTH, passing DM message."
                )

        except ModuleNotFoundError:
            send_aa_discordbot_notification(user.pk, message)
    else:
        logger.debug(
            "Aadiscordnotify is already active, passing notification sending to prevent multiple notifications"
        )


# This will send the contract notification to a selected webhook on discord
def send_message_to_discord_channel(
    webhook, message: dict, embed: bool = False
) -> None:
    logger.debug("Notifications: Starting to parse channel webhook notification")

    url = webhook  # webhook url, from here: https://i.imgur.com/f9XnAew.png

    # for all params, see https://discordapp.com/developers/docs/resources/webhook#execute-webhook
    data = {}

    # leave this out if you dont want an embed
    # for all params, see https://discordapp.com/developers/docs/resources/channel#embed-object

    contract = message["contract"]
    tracking = message["tracking"]

    data["embeds"] = [
        {
            "description": message["contract_items"],
            "title": message["title"],
            "color": message["color"],
            "fields": [
                {
                    "name": "Date issued",
                    "value": str(contract.date_issued),
                    "inline": True,
                },
                {
                    "name": "Issued from",
                    "value": str(message["assigned_from"]),
                    "inline": True,
                },
                {
                    "name": "Issued to",
                    "value": str(message["assigned_to"]),
                    "inline": True,
                },
                {
                    "name": "Location",
                    "value": str(contract.location_name),
                    "inline": True,
                },
                {
                    "name": "Tracking #",
                    "value": str(tracking.tracking_number),
                    "inline": True,
                },
                {
                    "name": "Volume",
                    "value": str(intcomma(int(contract.volume))) + " m3",
                    "inline": True,
                },
                {
                    "name": "Value",
                    "value": str(intcomma(int(contract.price))) + " ISK",
                    "inline": True,
                },
                {"name": "Notes", "value": str(message["notes"]), "inline": False},
            ],
        }
    ]

    result = requests.post(url, json=data)

    try:
        result.raise_for_status()
    except requests.exceptions.HTTPError as err:
        logger.error(err)
    else:
        logger.debug(
            "Payload delivered successfully, code {}.".format(result.status_code)
        )



# Generated by Django 4.0.6 on 2023-04-16 09:15

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0012_make_contract_id_unique"),
    ]

    operations = [
        migrations.AddField(
            model_name="program",
            name="discord_show_item_list",
            field=models.BooleanField(
                default=False,
                help_text="Determines if you want to show the contract items in the discord messages. This applies to both webhooks and direct messages.",
                verbose_name="Show list of items on discord message",
            ),
        ),
    ]



# Generated by Django 4.0.6 on 2022-11-06 17:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0008_alter_general_options_alter_program_refining_rate"),
    ]

    operations = [
        migrations.AddField(
            model_name="contract",
            name="location_name",
            field=models.CharField(max_length=128, null=True),
        ),
    ]



"""Delete contract duplicates."""

from django.db import migrations
from collections import Counter
import logging
from typing import Set

logger = logging.getLogger(__name__)


def forwards(apps, schema_editor):
    Contract = apps.get_model("buybackprogram", "Contract")
    duplicate_ids = _identify_duplicates(Contract)
    logger.debug("Found %d duplicate contract IDs", len(duplicate_ids))
    if not duplicate_ids:
        logger.info("No duplicate contracts to delete.")
        return
    for contract_id in duplicate_ids:
        _cleanup_contract_id(Contract, contract_id)

    print(f"Cleaned up approx. {len(duplicate_ids)} duplicate contracts.", end="")
    if remaining_duplicates := _identify_duplicates(Contract):
        logger.error(
            "Failed to clean up all duplicate contracts. "
            f"These duplicate contract IDs remain: {remaining_duplicates}"
        )
        print("ERROR: Failed to cleanup all duplicate contracts. Aborting.")
        exit(1)


def _identify_duplicates(Contract) -> Set[int]:
    ids = list(Contract.objects.values_list("contract_id", flat=True))
    ids_counted = Counter(ids)
    return {id for id, count in ids_counted.items() if count > 1}


def _cleanup_contract_id(Contract, contract_id):
    """Cleanup contracts with the same contract ID.

    We try to find one good contract to keep and then delete the others.
    """
    duplicates = Contract.objects.filter(contract_id=contract_id).order_by("id")
    logger.debug(
        "Trying to cleanup %d contracts with ID %d", duplicates.count(), contract_id
    )
    contracts_with_tracking = {"tracking__isnull": False, "no_tracking": False}
    contracts_without_tracking = {"tracking__isnull": True, "no_tracking": True}
    contracts_with_tracking_2 = {"tracking__isnull": True, "no_tracking": False}
    contracts_without_tracking_2 = {"tracking__isnull": False, "no_tracking": True}
    if duplicates.filter(**contracts_with_tracking).count() > 0:
        _reduce_contracts(duplicates, contracts_with_tracking)
    elif duplicates.filter(**contracts_without_tracking).count() > 0:
        _reduce_contracts(duplicates, contracts_without_tracking)
    elif duplicates.filter(**contracts_with_tracking_2).count() > 0:
        _reduce_contracts(duplicates, contracts_with_tracking_2)
    elif duplicates.filter(**contracts_without_tracking_2).count() > 0:
        _reduce_contracts(duplicates, contracts_without_tracking_2)


def _reduce_contracts(duplicates, params: dict):
    """Delete all contracts from duplicates except one chosen as primary."""
    primary_contract = duplicates.filter(**params).last()
    count, _ = duplicates.exclude(pk=primary_contract.pk).delete()
    if count > 0:
        logger.info(
            "Deleted %d duplicates for contract ID %d",
            count,
            primary_contract.contract_id,
        )
        logger.debug("Contract reduction success with params: %s", params)


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0010_contract_no_tracking_alter_tracking_tracking_number"),
    ]

    operations = [
        migrations.RunPython(forwards, migrations.RunPython.noop),
    ]



# Generated by Django 4.0.9 on 2023-04-05 19:18

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0011_delete_duplicated_contracts"),
    ]

    operations = [
        migrations.AlterField(
            model_name="contract",
            name="contract_id",
            field=models.IntegerField(unique=True),
        ),
    ]



from django.test import TestCase
from eveuniverse.models import EveType
from eveuniverse.tools.testdata import ModelSpec, create_testdata

from . import eveuniverse_test_data_filename


class CreateEveUniverseTestData(TestCase):
    def test_create_testdata(self):
        testdata_spec = [
            ModelSpec("EveSolarSystem", ids=[30000142, 30004984, 30001161, 30002537]),
            ModelSpec(
                "EveCategory",
                ids=[25],  # Asteroid
                include_children=True,
                enabled_sections=[
                    EveType.Section.TYPE_MATERIALS,
                    EveType.Section.DOGMAS,
                ],
            ),
        ]
        create_testdata(testdata_spec, eveuniverse_test_data_filename())



# Generated by Django 3.2.10 on 2022-01-08 14:35

import django.core.validators
import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import buybackprogram.models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("eveuniverse", "0005_type_materials_and_sections"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("eveonline", "0015_factions"),
        ("authentication", "0019_merge_20211026_0919"),
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="General",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    ("basic_access", "Can access this app and see own statics."),
                    (
                        "manage_programs",
                        "Can manage own buyback programs and see own program statics.",
                    ),
                    ("see_all_statics", "Can see all program statics."),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Contract",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("assignee_id", models.IntegerField()),
                ("availability", models.CharField(max_length=20)),
                ("contract_id", models.IntegerField()),
                ("date_completed", models.DateTimeField(null=True)),
                ("date_expired", models.DateTimeField(null=True)),
                ("date_issued", models.DateTimeField()),
                ("for_corporation", models.BooleanField()),
                ("issuer_corporation_id", models.IntegerField()),
                ("issuer_id", models.IntegerField()),
                ("start_location_id", models.BigIntegerField(null=True)),
                ("price", models.BigIntegerField()),
                ("status", models.CharField(max_length=30)),
                ("title", models.CharField(max_length=128)),
                ("volume", models.BigIntegerField()),
            ],
        ),
        migrations.CreateModel(
            name="Location",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="Structure name where the contracts are accepted at",
                        max_length=32,
                    ),
                ),
                (
                    "structure_id",
                    models.BigIntegerField(
                        blank=True,
                        default=None,
                        help_text="The ID for the structure you wish to accept the contracts at. If left empty the program statistics page will not track if the contract is actually made at the correct structure or not. To get the ID for the structure see readme for getting structure IDs",
                        null=True,
                        verbose_name="Ingame unique ID for structure",
                    ),
                ),
                (
                    "eve_solar_system",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        help_text="System where the buyback structure is located",
                        null=True,
                        on_delete=django.db.models.deletion.SET_DEFAULT,
                        related_name="+",
                        to="eveuniverse.evesolarsystem",
                        verbose_name="Solar system",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Owner",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "character",
                    models.ForeignKey(
                        help_text="Character used for retrieving info",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to="authentication.characterownership",
                    ),
                ),
                (
                    "corporation",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="eveonline.evecorporationinfo",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User that manages the program",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Program",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "is_corporation",
                    models.BooleanField(
                        default=False,
                        help_text="If we should use the corporation of the manager as the contract receiver instead of the character.",
                    ),
                ),
                (
                    "tax",
                    models.IntegerField(
                        default=0,
                        help_text="A default tax rate in this program that is applied on all items.",
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                        verbose_name="Default tax",
                    ),
                ),
                (
                    "hauling_fuel_cost",
                    models.IntegerField(
                        default=0,
                        help_text="ISK per m³ that will be removed from the buy price ie. to cover jump freighet fuel costs. <b>Should not be used with price dencity modifier</b>",
                        verbose_name="Hauling fuel cost per m³",
                    ),
                ),
                (
                    "price_dencity_modifier",
                    models.BooleanField(
                        default=False,
                        help_text="Should we modify buy prices for items with high volume and low value ie. T1 industrial hulls. <b>Should not be used with hauling fuel cost</b>",
                        verbose_name="Price density modifier",
                    ),
                ),
                (
                    "price_dencity_treshold",
                    models.IntegerField(
                        default=0,
                        help_text="At what ISK/m3 do we start to apply the low isk dencity tax. Tritanium is 500 ISK/m³ @ 5 ISK per unit price. PLEX is 14,5Trillion ISK/m³ @2.9M per unit price.",
                        null=True,
                        verbose_name="Price density threshold",
                    ),
                ),
                (
                    "price_dencity_tax",
                    models.IntegerField(
                        default=0,
                        help_text="How much tax do we apply on the low isk density items.",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                        verbose_name="Price density tax",
                    ),
                ),
                (
                    "allow_all_items",
                    models.BooleanField(
                        default=True,
                        help_text="If true all items are accepted to the buyback program. You can set extra taxes or disallow individual items from the program item section. If set to false you need to add each accepted item into the program item section. Blueprints are not included in all items.",
                    ),
                ),
                (
                    "use_refined_value",
                    models.BooleanField(
                        default=False,
                        help_text="Take refined value into account when calculating prices for ore, ice and moon goo",
                        verbose_name="Ore: Use refined value",
                    ),
                ),
                (
                    "use_compressed_value",
                    models.BooleanField(
                        default=False,
                        help_text="Take compressed value into account when calculating prices for ore, ice and moon goo",
                        verbose_name="Ore: Use compressed value",
                    ),
                ),
                (
                    "use_raw_ore_value",
                    models.BooleanField(
                        default=True,
                        help_text="Take raw ore value into account when calculating prices for ore, ice and moon goo",
                        verbose_name="Ore: Use raw value",
                    ),
                ),
                (
                    "allow_unpacked_items",
                    models.BooleanField(
                        default=False,
                        help_text="Do you want to allow unpacked items in this program such as assembled ship hulls?",
                        verbose_name="Allow unpacked items",
                    ),
                ),
                (
                    "refining_rate",
                    models.IntegerField(
                        default=0,
                        help_text="Refining rate to be used if ore refined value is active",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(0),
                        ],
                        verbose_name="Refining rate",
                    ),
                ),
                (
                    "blue_loot_npc_price",
                    models.BooleanField(
                        default=False,
                        help_text="Use NPC price as value for blue loot",
                        verbose_name="NPC price for: Sleeper Components",
                    ),
                ),
                (
                    "red_loot_npc_price",
                    models.BooleanField(
                        default=False,
                        help_text="Use NPC price as value for red loot",
                        verbose_name="NPC price for: Triglavian Survey Database",
                    ),
                ),
                (
                    "ope_npc_price",
                    models.BooleanField(
                        default=False,
                        help_text="Use NPC price as value for OPEs",
                        verbose_name="NPC price for: Overseer's Personal Effects",
                    ),
                ),
                (
                    "discord_dm_notification",
                    models.BooleanField(
                        default=False,
                        help_text="Check if you want to receive a direct message notification each time a new contract is created. <b>Requires aa-discordbot app or discordproxy app to work</b>",
                        verbose_name="Discord direct messages for new contracts",
                    ),
                ),
                (
                    "discord_channel_notification",
                    models.BigIntegerField(
                        blank=True,
                        help_text="Check if you want to feed new contracts to a discord channel. <b>Requires aa-discordbot app or discordproxy app to work</b>",
                        null=True,
                        verbose_name="Discord channel ID for notifications",
                    ),
                ),
                (
                    "location",
                    models.ForeignKey(
                        help_text="The location where contracts should be created at.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="buybackprogram.location",
                    ),
                ),
                (
                    "owner",
                    models.ForeignKey(
                        help_text="Character that is used to manage this program.",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to="buybackprogram.owner",
                        verbose_name="Manager",
                    ),
                ),
                (
                    "restricted_to_group",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The group(s) that will be able to see this buyback program. If none is selected program is open for all.",
                        related_name="buybackprogram_require_groups",
                        to="auth.Group",
                    ),
                ),
                (
                    "restricted_to_state",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The state(s) that will be able to see this buyback program. If none is selected program is open for all.",
                        related_name="buybackprogram_require_states",
                        to="authentication.State",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="Tracking",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("value", models.BigIntegerField()),
                ("taxes", models.BigIntegerField()),
                ("hauling_cost", models.BigIntegerField()),
                ("donation", models.BigIntegerField(blank=True, null=True)),
                ("net_price", models.BigIntegerField()),
                ("tracking_number", models.CharField(max_length=20)),
                (
                    "issuer_user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="+",
                        to="buybackprogram.program",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="UserSettings",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("disable_notifications", models.BooleanField(default=False)),
                (
                    "user",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=models.SET(buybackprogram.models.get_sentinel_user),
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "User Settings",
                "verbose_name_plural": "User Settings",
                "default_permissions": (),
            },
        ),
        migrations.CreateModel(
            name="TrackingItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy_value", models.BigIntegerField()),
                ("quantity", models.IntegerField()),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="Item type information",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "tracking",
                    models.ForeignKey(
                        help_text="What tracking do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.tracking",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="location",
            name="owner",
            field=models.ForeignKey(
                help_text="Player managing this location",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="buybackprogram.owner",
                verbose_name="Manager",
            ),
        ),
        migrations.CreateModel(
            name="ItemPrices",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("buy", models.BigIntegerField()),
                ("sell", models.BigIntegerField()),
                ("updated", models.DateTimeField()),
                (
                    "eve_type",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ContractItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("quantity", models.IntegerField()),
                (
                    "contract",
                    models.ForeignKey(
                        help_text="What contract do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.contract",
                    ),
                ),
                (
                    "eve_type",
                    models.ForeignKey(
                        help_text="Item type information",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="ProgramItem",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "item_tax",
                    models.IntegerField(
                        default=0,
                        help_text="How much do you want to adjust the default tax on this item. Can be a positive or a negative value.",
                        null=True,
                        validators=[
                            django.core.validators.MaxValueValidator(100),
                            django.core.validators.MinValueValidator(-100),
                        ],
                        verbose_name="Item tax adjustment",
                    ),
                ),
                (
                    "disallow_item",
                    models.BooleanField(
                        default=False,
                        help_text="You can disallow an item from a buyback location. It will return 0 price if disallowed.",
                        verbose_name="Disallow item in program",
                    ),
                ),
                (
                    "item_type",
                    models.ForeignKey(
                        help_text="Select item for special tax",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="eveuniverse.evetype",
                    ),
                ),
                (
                    "program",
                    models.ForeignKey(
                        help_text="What program do these items belong to",
                        on_delete=django.db.models.deletion.CASCADE,
                        to="buybackprogram.program",
                    ),
                ),
            ],
            options={
                "default_permissions": (),
                "unique_together": {("program", "item_type")},
            },
        ),
    ]



# Generated by Django 4.0.6 on 2023-03-27 13:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0009_contract_location_name"),
    ]

    operations = [
        migrations.AddField(
            model_name="contract",
            name="no_tracking",
            field=models.BooleanField(default=False),
        ),
        migrations.AlterField(
            model_name="tracking",
            name="tracking_number",
            field=models.CharField(max_length=32),
        ),
    ]



import unittest
from importlib import import_module

from buybackprogram.models import Contract
from django.apps import apps
from django.db import connection
from django.test import TestCase
from app_utils.testing import set_test_logger
from .testdata.factories import ContractFactory, ProgramFactory, TrackingFactory
from .testdata.load_eveuniverse import load_eveuniverse

data_migration = import_module(
    "buybackprogram.migrations.0011_delete_duplicated_contracts"
)

set_test_logger(data_migration.logger.name, __file__)


def all_contracts() -> set:
    return set(Contract.objects.all())


@unittest.skip  # Test no longer works after migration 0012 is applied
class TestDataMigration(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        cls.program = ProgramFactory()

    def test_should_do_nothing_when_no_duplicates(self):
        # given
        contract = ContractFactory()
        # when
        data_migration.forwards(apps, connection.schema_editor())
        # then
        self.assertSetEqual(all_contracts(), {contract})

    def test_should_cleanup_variant_1(self):
        # given
        contract_1 = ContractFactory(contract_id=1)
        contract_2a = ContractFactory(contract_id=2, no_tracking=False)
        contract_2b = ContractFactory(contract_id=2, no_tracking=False)
        ContractFactory(contract_id=2, no_tracking=False)
        ContractFactory(contract_id=2, no_tracking=True)
        TrackingFactory(program=self.program, contract=contract_2a)
        TrackingFactory(program=self.program, contract=contract_2b)
        # when
        data_migration.forwards(apps, connection.schema_editor())
        # then
        self.assertSetEqual(all_contracts(), {contract_1, contract_2b})

    def test_should_cleanup_variant_2(self):
        # given
        contract_1 = ContractFactory(contract_id=1)
        ContractFactory(contract_id=2, no_tracking=True)
        contract_2b = ContractFactory(contract_id=2, no_tracking=True)
        contract_2c = ContractFactory(contract_id=2, no_tracking=True)
        ContractFactory(contract_id=2, no_tracking=False)
        TrackingFactory(program=self.program, contract=contract_2c)
        # when
        data_migration.forwards(apps, connection.schema_editor())
        # then
        self.assertSetEqual(all_contracts(), {contract_1, contract_2b})

    def test_should_cleanup_variant_3(self):
        # given
        contract_1 = ContractFactory(contract_id=1)
        ContractFactory(contract_id=2, no_tracking=False)
        contract_2b = ContractFactory(contract_id=2, no_tracking=False)
        contract_2c = ContractFactory(contract_id=2, no_tracking=True)
        TrackingFactory(program=self.program, contract=contract_2c)
        # when
        data_migration.forwards(apps, connection.schema_editor())
        # then
        self.assertSetEqual(all_contracts(), {contract_1, contract_2b})

    def test_should_cleanup_variant_4(self):
        # given
        contract_1 = ContractFactory(contract_id=1)
        contract_2a = ContractFactory(contract_id=2, no_tracking=True)
        contract_2b = ContractFactory(contract_id=2, no_tracking=True)
        TrackingFactory(program=self.program, contract=contract_2a)
        TrackingFactory(program=self.program, contract=contract_2b)
        # when
        data_migration.forwards(apps, connection.schema_editor())
        # then
        self.assertSetEqual(all_contracts(), {contract_1, contract_2b})



# Generated by Django 4.0.6 on 2022-09-20 14:45

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0007_program_expiration"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app and see own statics."),
                    (
                        "manage_programs",
                        "Can manage own buyback programs and see own program statics.",
                    ),
                    (
                        "see_leaderboard",
                        "Can see leaderboards for all available programs.",
                    ),
                    (
                        "see_performance",
                        "Can see performance for all available programs.",
                    ),
                    ("see_all_statics", "Can see all program statics."),
                ),
            },
        ),
        migrations.AlterField(
            model_name="program",
            name="refining_rate",
            field=models.DecimalField(
                decimal_places=2,
                default=0,
                help_text="Refining rate to be used if ore refined value is active",
                max_digits=5,
                null=True,
                validators=[
                    django.core.validators.MaxValueValidator(100),
                    django.core.validators.MinValueValidator(0),
                ],
                verbose_name="Refining rate",
            ),
        ),
    ]



# Generated by Django 4.0.4 on 2022-06-18 11:53

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0005_program_compression_price_dencity_modifier"),
    ]

    operations = [
        migrations.AddField(
            model_name="program",
            name="bonds_npc_price",
            field=models.BooleanField(
                default=False,
                help_text="Use NPC price as value for SCC Encrypted Bond",
                verbose_name="NPC price for: Bounty Encrypted Bonds",
            ),
        ),
        migrations.AlterField(
            model_name="program",
            name="discord_channel_notification",
            field=models.CharField(
                blank=True,
                help_text="Discord webhook to send contract notifications to.",
                max_length=256,
                null=True,
                verbose_name="Discord webhook for notifications",
            ),
        ),
        migrations.AlterField(
            model_name="program",
            name="location",
            field=models.ManyToManyField(
                help_text="The location where contracts should be created at.",
                related_name="+",
                to="buybackprogram.location",
            ),
        ),
    ]



import datetime as dt

import factory
import factory.fuzzy
from faker import Faker

from django.utils.timezone import now
from eveuniverse.models import EveEntity, EveSolarSystem, EveType

from app_utils.testdata_factories import (
    EveAllianceInfoFactory,
    EveCharacterFactory,
    EveCorporationInfoFactory,
    UserMainFactory,
)

from buybackprogram.models import (
    Contract,
    ContractItem,
    ContractNotification,
    ItemPrices,
    Location,
    Owner,
    Program,
    ProgramItem,
    Tracking,
    TrackingItem,
    UserSettings,
)

fake = Faker()
Faker.seed(0)


def random_eve_type_id() -> int:
    ids = EveType.objects.filter(published=True).values_list("id", flat=True)
    if not ids:
        return None
    return factory.fuzzy.FuzzyChoice(ids).fuzz()


def random_eve_type() -> EveType:
    eve_type_did = random_eve_type_id()
    if not eve_type_did:
        return None
    return EveType.objects.get(id=eve_type_did)


# TODO: Move to app_utils
class EveEntityFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = EveEntity
        django_get_or_create = ("id", "name")

    category = EveEntity.CATEGORY_CHARACTER

    @factory.lazy_attribute
    def id(self):
        if self.category == EveEntity.CATEGORY_CHARACTER:
            obj = EveCharacterFactory()
            return obj.character_id
        if self.category == EveEntity.CATEGORY_CORPORATION:
            obj = EveCorporationInfoFactory()
            return obj.corporation_id
        if self.category == EveEntity.CATEGORY_ALLIANCE:
            obj = EveAllianceInfoFactory()
            return obj.alliance_id
        raise NotImplementedError(f"Unknown category: {self.category}")


class EveEntityCharacterFactory(EveEntityFactory):
    name = factory.Faker("name")
    category = EveEntity.CATEGORY_CHARACTER


class EveEntityCorporationFactory(EveEntityFactory):
    name = factory.Faker("company")
    category = EveEntity.CATEGORY_CORPORATION


class EveEntityAllianceFactory(EveEntityFactory):
    name = factory.Faker("company")
    category = EveEntity.CATEGORY_ALLIANCE


class UserProjectManagerFactory(UserMainFactory):
    main_character__scopes = [
        "esi-contracts.read_character_contracts.v1",
        "esi-contracts.read_corporation_contracts.v1",
        "esi-universe.read_structures.v1",
    ]
    permissions__ = ["buybackprogram.basic_access", "buybackprogram.manage_programs"]


class UserIssuerFactory(UserMainFactory):
    main_character__scopes = [
        "esi-contracts.read_character_contracts.v1",
        "esi-contracts.read_corporation_contracts.v1",
        "esi-universe.read_structures.v1",
    ]
    permissions__ = ["buybackprogram.basic_access"]


class OwnerFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Owner

    user = factory.SubFactory(UserProjectManagerFactory)
    character = factory.lazy_attribute(
        lambda o: o.user.profile.main_character.character_ownership
    )

    @factory.lazy_attribute
    def corporation(self):
        return EveCorporationInfoFactory(
            corporation_id=self.user.profile.main_character.corporation_id
        )


class LocationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Location

    name = factory.Faker("word")
    owner = factory.SubFactory(OwnerFactory)

    @factory.lazy_attribute
    def eve_solar_system(self):
        obj = EveSolarSystem.objects.order_by("?").first()
        if not obj:
            raise RuntimeError("No EveSolarSystem found for LocationFactory.")
        return obj


class ProgramFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Program

    name = factory.Faker("sentence")
    owner = factory.SubFactory(OwnerFactory)

    @factory.post_generation
    def location(self, create, extracted, **kwargs):
        if not create:
            return

        if extracted:
            for obj in extracted:
                self.location.add(obj)
        else:
            obj = LocationFactory()
            self.location.add(obj)


class ProgramItemFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ProgramItem

    program = factory.SubFactory(ProgramFactory)

    @factory.lazy_attribute
    def item_type(self):
        obj = random_eve_type()
        if not obj:
            raise RuntimeError("No EveType found for ProgramItemFactory.")
        return obj


class ItemPricesFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ItemPrices

    buy = factory.fuzzy.FuzzyInteger(1, 10_000_000)
    sell = factory.fuzzy.FuzzyInteger(1, 10_000_000)
    updated = factory.lazy_attribute(lambda o: now())

    @factory.lazy_attribute
    def eve_type(self):
        obj = random_eve_type()
        if not obj:
            raise RuntimeError("No EveType found for ProgramItemFactory.")
        return obj


class ContractFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Contract

    assignee_id = factory.Sequence(lambda n: 90_000 + n)
    availability = "public"
    contract_id = factory.Sequence(lambda n: 1_000_000_000 + n)
    date_issued = factory.fuzzy.FuzzyDateTime(now() - dt.timedelta(days=3))
    for_corporation = False
    issuer_corporation_id = factory.Sequence(lambda n: 99_000 + n)
    issuer_id = factory.Sequence(lambda n: 95_000 + n)
    price = factory.fuzzy.FuzzyInteger(90_000, 99_000)
    status = "outstanding"
    title = factory.Faker("sentence")
    volume = factory.fuzzy.FuzzyInteger(10, 320)


class EsiContractFactory(factory.DictFactory):
    """Contract dictionary returned from ESI endpoint."""

    # acceptor_id
    assignee_id = factory.Sequence(lambda n: 90_000 + n)
    availability = "public"
    contract_id = factory.Sequence(lambda n: 1_000_000_000 + n)
    date_issued = factory.fuzzy.FuzzyDateTime(now() - dt.timedelta(days=3))
    for_corporation = False
    issuer_corporation_id = factory.Sequence(lambda n: 99_000 + n)
    issuer_id = factory.Sequence(lambda n: 95_000 + n)
    price = factory.fuzzy.FuzzyInteger(90_000, 99_000)
    status = "outstanding"
    start_location_id = 60003760
    title = factory.Faker("sentence")
    type = "item_exchange"
    volume = factory.fuzzy.FuzzyInteger(10, 320)

    @factory.lazy_attribute
    def date_expired(self):
        return factory.fuzzy.FuzzyDateTime(
            self.date_issued, self.date_issued + dt.timedelta(days=10)
        ).fuzz()

    @factory.lazy_attribute
    def date_completed(self):
        return factory.fuzzy.FuzzyDateTime(self.date_issued, self.date_expired).fuzz()


class EsiContractItemFactory(factory.DictFactory):
    """Contract item dictionary returned from ESI endpoint."""

    is_included = True
    is_singleton = False
    quantity = factory.fuzzy.FuzzyInteger(1, 999)
    record_id = factory.Sequence(lambda n: 1 + n)

    @factory.lazy_attribute
    def type_id(self):
        id = random_eve_type_id()
        if not id:
            raise RuntimeError("No EveType found for EsiContractItemFactory.")
        return id


class ContractItemFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ContractItem

    contract = factory.SubFactory(ContractFactory)
    quantity = factory.fuzzy.FuzzyInteger(1, 999)

    @factory.lazy_attribute
    def eve_type(self):
        obj = random_eve_type()
        if not obj:
            raise RuntimeError("No EveType found for ProgramItemFactory.")
        return obj


class ContractNotificationFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = ContractNotification

    contract = factory.SubFactory(ContractFactory)
    icon = factory.Faker("url")
    color = factory.fuzzy.FuzzyChoice(["green", "orange", "red"])
    message = factory.Faker("sentence")


class TrackingFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = Tracking

    program = factory.SubFactory(ProgramFactory)
    issuer_user = factory.SubFactory(UserIssuerFactory)
    value = factory.fuzzy.FuzzyInteger(1_000_000, 100_000_000)
    taxes = factory.fuzzy.FuzzyInteger(1_000_000, 5_000_000)
    net_price = factory.fuzzy.FuzzyInteger(1_000_000, 100_000_000)
    hauling_cost = factory.fuzzy.FuzzyInteger(1_000_000, 5_000_000)
    tracking_number = 5  # what is this?
    created_at = factory.fuzzy.FuzzyDateTime(now() - dt.timedelta(days=3))

    @factory.lazy_attribute
    def tracking_number(self):
        return "aa-bbp-" + fake.uuid4()[:6].upper()


class TrackingItemFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TrackingItem

    tracking = factory.SubFactory(TrackingFactory)
    buy_value = factory.fuzzy.FuzzyInteger(1_000_000, 100_000_000)
    quantity = factory.fuzzy.FuzzyInteger(1, 999)

    @factory.lazy_attribute
    def eve_type(self):
        obj = random_eve_type()
        if not obj:
            raise RuntimeError("No EveType found for ProgramItemFactory.")
        return obj


class UserSettingsFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = UserSettings

    user = factory.SubFactory(UserIssuerFactory)



from django.conf import settings
import re

from .utils import clean_setting

# put your app settings here


def get_site_url():  # regex sso url
    regex = r"^(.+)\/s.+"
    matches = re.finditer(regex, settings.ESI_SSO_CALLBACK_URL, re.MULTILINE)
    url = "http://"

    for m in matches:
        url = m.groups()[0]  # first match

    return url


# Hard timeout for tasks in seconds to reduce task accumulation during outages
BUYBACKPROGRAM_TASKS_TIME_LIMIT = clean_setting("BUYBACKPROGRAM_TASKS_TIME_LIMIT", 7200)

# Warning limit for Jita price updates if prices have not been updated
BUYBACKPROGRAM_PRICE_AGE_WARNING_LIMIT = clean_setting(
    "BUYBACKPROGRAM_PRICE_AGE_WARNING_LIMIT", 48
)

# Unused tracking purge limit
BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT = clean_setting(
    "BUYBACKPROGRAM_UNUSED_TRACKING_PURGE_LIMIT", 48
)

BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS = clean_setting(
    "BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS", True
)

# Tracking number tag
BUYBACKPROGRAM_TRACKING_PREFILL = clean_setting(
    "BUYBACKPROGRAM_TRACKING_PREFILL", "aa-bbp"
)

BUYBACKPROGRAM_PRICE_SOURCE_ID = clean_setting(
    "BUYBACKPROGRAM_PRICE_SOURCE_ID", 60003760
)

BUYBACKPROGRAM_PRICE_SOURCE_NAME = clean_setting(
    "BUYBACKPROGRAM_PRICE_SOURCE_NAME", "Jita"
)

BUYBACKPROGRAM_PRICE_METHOD = clean_setting("BUYBACKPROGRAM_PRICE_METHOD", "Fuzzwork")

BUYBACKPROGRAM_PRICE_JANICE_API_KEY = clean_setting(
    "BUYBACKPROGRAM_PRICE_JANICE_API_KEY", ""
)


def allianceauth_discordbot_active():
    """
    check if allianceauth-dicordbot is installed and active
    :return:
    """
    return "aadiscordbot" in settings.INSTALLED_APPS


def aa_discordnotify_active():
    """
    check if allianceauth-dicordbot is installed and active
    :return:
    """
    return "discordnotify" in settings.INSTALLED_APPS



import random
import uuid
from datetime import timedelta

from django.contrib.auth.models import User
from django.core.management.base import BaseCommand
from django.utils import timezone
from eveuniverse.models import EveType

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.models import Contract, Program, Tracking, TrackingItem

logger = get_extension_logger(__name__)


def get_input(text):
    """wrapped input to enable unit testing / patching"""
    return input(text)


class Command(BaseCommand):
    help = "Links tracking objects with old contracts after 1.2.0 update"

    def handle(self, *args, **options):
        self.stdout.write(
            "\033[91mThis command is only for development purposes. It will generate an endless amount of dummy contracts until terminated. Do not use in production enviroment!\033[91m\033[0m"
        )
        user_input = get_input(
            "Are you sure you want to proceed and run the command? (y/N)?"
        )
        if user_input.lower() == "y":
            x = 1

            while True:
                x = x

                net_price = random.randint(1000000, 100000000)

                donation = random.randint(100000, 10000000)

                time = timezone.now() - timedelta(x)

                contract_id = x

                program = Program.objects.first()

                user = User.objects.first()

                tracking_number = "aa-bbp-1234-" + uuid.uuid4().hex[:6].upper()

                # Create or update the found contract
                obj, created = Contract.objects.update_or_create(
                    contract_id=contract_id,
                    defaults={
                        "contract_id": contract_id,
                        "assignee_id": 93337205,
                        "availability": "personal",
                        "date_completed": time,
                        "date_expired": time,
                        "date_issued": time,
                        "for_corporation": 0,
                        "issuer_corporation_id": 93337205,
                        "issuer_id": 93337205,
                        "start_location_id": 123,
                        "price": net_price,
                        "status": "finished",
                        "title": tracking_number,
                        "volume": 100,
                    },
                )

                tracking = Tracking(
                    program=program,
                    issuer_user=user,
                    value=10000000,
                    taxes=500000,
                    hauling_cost=0,
                    donation=donation,
                    contract_id=obj.id,
                    net_price=net_price,
                    tracking_number=tracking_number,
                    created_at=time,
                )

                tracking.save()

                objs = []

                for n in range(0, 3):
                    item_type = EveType.objects.order_by("?").first()

                    tracking_item = TrackingItem(
                        tracking=tracking,
                        eve_type=item_type,
                        buy_value=10,
                        quantity=2,
                    )

                    objs.append(tracking_item)

                TrackingItem.objects.bulk_create(objs)

                x = x + 1

                logger.debug("Generated contract %s" % contract_id)
            else:
                self.stdout.write("Aborted")



from unittest.mock import patch

from django.db import IntegrityError
from django.test import TestCase
from eveuniverse.models import EveType

from app_utils.esi import EsiStatus
from app_utils.esi_testing import EsiClientStub, EsiEndpoint
from app_utils.testing import NoSocketsTestCase

from buybackprogram.models import Contract

from .testdata.factories import (
    ContractFactory,
    ContractItemFactory,
    ContractNotificationFactory,
    EsiContractFactory,
    EsiContractItemFactory,
    EveEntityCharacterFactory,
    ItemPricesFactory,
    LocationFactory,
    OwnerFactory,
    ProgramFactory,
    ProgramItemFactory,
    TrackingFactory,
    TrackingItemFactory,
    UserProjectManagerFactory,
    UserSettingsFactory,
)
from .testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "buybackprogram.models"


class TestOwners(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        cls.project_manager = UserProjectManagerFactory()

    def test_should_have_str_method(self):
        # given
        obj = OwnerFactory(user=self.project_manager)
        # when/then
        self.assertIsInstance(str(obj), str)


@patch(MODULE_PATH + ".send_user_notification")
@patch(MODULE_PATH + ".Owner._get_location_name")
@patch(MODULE_PATH + ".esi")
class TestOwnersUpdateContractFromEsi(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        cls.project_manager = UserProjectManagerFactory()

    def test_should_create_new_contract(
        self, mock_esi, mock_get_location_name, mock_send_user_notification
    ):
        # given
        tracking_number = "AB1"
        eve_type = EveType.objects.get(name="Pyerite")
        owner = OwnerFactory(user=self.project_manager)
        owner_character_id = owner.character.character.character_id
        owner_corporation_id = owner.corporation.corporation_id
        issuer = EveEntityCharacterFactory()
        esi_contract = EsiContractFactory(title=tracking_number, issuer_id=issuer.id)
        esi_contract_item = EsiContractItemFactory(type_id=eve_type.id)
        endpoints = [
            EsiEndpoint(
                "Contracts",
                "get_characters_character_id_contracts",
                "character_id",
                needs_token=True,
                data={str(owner_character_id): [esi_contract]},
            ),
            EsiEndpoint(
                "Contracts",
                "get_characters_character_id_contracts_contract_id_items",
                ("character_id", "contract_id"),
                needs_token=True,
                data={
                    str(owner_character_id): {
                        str(esi_contract["contract_id"]): [esi_contract_item]
                    }
                },
            ),
            EsiEndpoint(
                "Contracts",
                "get_corporations_corporation_id_contracts",
                "corporation_id",
                needs_token=True,
                data={str(owner_corporation_id): []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        program = ProgramFactory(owner=owner)
        tracking = TrackingFactory(program=program, tracking_number=tracking_number)
        TrackingItemFactory(tracking=tracking, eve_type=eve_type)
        mock_get_location_name.return_value = "Unknown"
        # when
        owner.update_contracts_esi()
        # then
        self.assertEqual(Contract.objects.count(), 1)
        contract = Contract.objects.first()
        self.assertEqual(contract.contract_id, esi_contract["contract_id"])
        self.assertEqual(contract.title, esi_contract["title"])
        self.assertEqual(contract.contractitem_set.count(), 1)
        item = contract.contractitem_set.first()
        self.assertEqual(item.eve_type_id, esi_contract_item["type_id"])
        self.assertEqual(item.quantity, esi_contract_item["quantity"])

    def test_should_update_existing_contract(
        self, mock_esi, mock_get_location_name, mock_send_user_notification
    ):
        # given
        tracking_number = "AB1"
        eve_type = EveType.objects.get(name="Pyerite")
        owner = OwnerFactory(user=self.project_manager)
        owner_character_id = owner.character.character.character_id
        owner_corporation_id = owner.corporation.corporation_id
        issuer = EveEntityCharacterFactory()
        contract = ContractFactory(title=tracking_number, issuer_id=issuer.id)
        esi_contract = EsiContractFactory(
            contract_id=contract.contract_id,
            title=tracking_number,
            issuer_id=issuer.id,
            status="finished",
        )
        endpoints = [
            EsiEndpoint(
                "Contracts",
                "get_characters_character_id_contracts",
                "character_id",
                needs_token=True,
                data={str(owner_character_id): [esi_contract]},
            ),
            EsiEndpoint(
                "Contracts",
                "get_corporations_corporation_id_contracts",
                "corporation_id",
                needs_token=True,
                data={str(owner_corporation_id): []},
            ),
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        program = ProgramFactory(owner=owner)
        tracking = TrackingFactory(program=program, tracking_number=tracking_number)
        TrackingItemFactory(tracking=tracking, eve_type=eve_type)
        UserSettingsFactory(user=tracking.issuer_user)
        mock_get_location_name.return_value = "Unknown"
        # when
        owner.update_contracts_esi()
        # then
        contract.refresh_from_db()
        self.assertEqual(contract.status, "finished")
        self.assertTrue(mock_send_user_notification.called)


@patch(MODULE_PATH + ".EveEntity.objects.resolve_name", spec=True)
@patch(MODULE_PATH + ".fetch_esi_status")
@patch(MODULE_PATH + ".esi")
class TestOwnerGetLocationName(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()
        cls.owner = OwnerFactory()

    def test_should_return_name_for_station(
        self, mock_esi, mock_fetch_esi_status, mock_resolve_name
    ):
        # given
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 99)
        mock_resolve_name.return_value = "Jita 4-4"
        # when
        result = self.owner._get_location_name(60003760)
        # then
        self.assertEqual(result, "Jita 4-4")

    def test_should_return_name_for_structure(
        self, mock_esi, mock_fetch_esi_status, mock_resolve_name
    ):
        # given
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 99)
        endpoints = [
            EsiEndpoint(
                "Universe",
                "get_universe_structures_structure_id",
                "structure_id",
                needs_token=True,
                data={"100000001": {"name": "My structure"}},
            )
        ]
        mock_esi.client = EsiClientStub.create_from_endpoints(endpoints)
        # when
        result = self.owner._get_location_name(100000001)
        # then
        self.assertEqual(result, "My structure")

    def test_should_return_unknown_when_esi_is_down(
        self, mock_esi, mock_fetch_esi_status, mock_resolve_name
    ):
        # given
        mock_fetch_esi_status.return_value = EsiStatus(False, 99, 99)
        mock_resolve_name.return_value = "Jita 4-4"
        # when
        result = self.owner._get_location_name(60003760)
        # then
        self.assertEqual(result, "Unknown")

    def test_should_return_unknown_when_structure_endpoints_returns_error(
        self, mock_esi, mock_fetch_esi_status, mock_resolve_name
    ):
        # given
        mock_fetch_esi_status.return_value = EsiStatus(True, 99, 99)
        mock = mock_esi.client.Universe
        mock.get_universe_structures_structure_id.return_value.result.side_effect = (
            OSError
        )
        # when
        result = self.owner._get_location_name(100000001)
        # then
        self.assertEqual(result, "Unknown")


class TestLocations(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = LocationFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestProgram(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = ProgramFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestProgramItem(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = ProgramItemFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestItemPrices(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = ItemPricesFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestContracts(TestCase):
    def test_should_have_str_method(self):
        # given
        obj = ContractFactory()
        # when/then
        self.assertIsInstance(str(obj), str)

    def test_contract_id_should_be_unique(self):
        # given
        contract = ContractFactory()
        # when/then
        with self.assertRaises(IntegrityError):
            ContractFactory(contract_id=contract.contract_id)


class TestContractItems(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = ContractItemFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestContractNotifications(TestCase):
    def test_should_have_str_method(self):
        # given
        obj = ContractNotificationFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestTrackings(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = TrackingFactory()
        # when/then
        self.assertIsInstance(str(obj), str)


class TestTrackingItems(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_should_have_str_method(self):
        # given
        obj = TrackingItemFactory()
        # when/then
        self.assertIsInstance(str(obj), str)



from django.core.management.base import BaseCommand
from django.db import IntegrityError
from django.utils import timezone
from eveuniverse.models import EveMarketPrice, EveType

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.app_settings import (
    BUYBACKPROGRAM_PRICE_METHOD,
    BUYBACKPROGRAM_PRICE_SOURCE_ID,
    BUYBACKPROGRAM_PRICE_SOURCE_NAME,
)
from buybackprogram.models import ItemPrices
from buybackprogram.tasks import get_bulk_prices, valid_janice_api_key

logger = get_extension_logger(__name__)


class Command(BaseCommand):
    help = (
        "Preloads price data required for the buyback program from Fuzzwork market API"
    )

    def handle(self, *args, **options):
        item_count = 0
        type_ids = []
        market_data = {}
        api_up = True

        # Get all type ids
        typeids = EveType.objects.values_list("id", flat=True).filter(published=True)

        if BUYBACKPROGRAM_PRICE_METHOD == "Fuzzwork":
            print(
                "Price setup starting for %s items from Fuzzworks API from station id %s (%s), this may take up to 30 seconds..."
                % (
                    len(typeids),
                    BUYBACKPROGRAM_PRICE_SOURCE_ID,
                    BUYBACKPROGRAM_PRICE_SOURCE_NAME,
                )
            )
        elif BUYBACKPROGRAM_PRICE_METHOD == "Janice":
            if valid_janice_api_key():
                print(
                    "Price setup starting for %s items from Janice API for Jita 4-4, this may take up to 30 seconds..."
                    % (len(typeids))
                )
            else:
                print(
                    "\033[91mPrice setup failed for Janice, invalid API key! Provide a working key or change price source to Fuzzwork\033[91m\033[0m"
                )

                api_up = False

        else:
            return (
                "Unknown pricing method: '%s', skipping" % BUYBACKPROGRAM_PRICE_METHOD
            )

        if api_up:
            # Build suitable bulks to fetch prices from API
            for item in typeids:
                type_ids.append(item)

                if len(type_ids) == 1000:
                    market_data.update(get_bulk_prices(type_ids))
                    type_ids.clear()

            # Get leftover data from the bulk
            market_data.update(get_bulk_prices(type_ids))

            objs = []

            for key, value in market_data.items():
                item_count += 1

                item = ItemPrices(
                    eve_type_id=key,
                    buy=int(float(value["buy"]["max"])),
                    sell=int(float(value["sell"]["min"])),
                    updated=timezone.now(),
                )

                objs.append(item)
            try:
                ItemPrices.objects.bulk_create(objs)

                print("Succesfully setup %s prices." % item_count)
            except IntegrityError:
                print(
                    "Error: Prices already loaded into database, did you mean to run task.update_all_prices instead?"
                )

                delete_arg = input("Would you like to delete current prices? (y/n): ")

                if delete_arg == "y":
                    ItemPrices.objects.all().delete()
                    return "All price data removed from database. Run the command again to populate the price data."
                else:
                    return "No changes done to price table."
            else:
                print("Starting to update NPC market prices for all fetched items...")

                EveMarketPrice.objects.update_from_esi()

                logger.debug("Updated all eveuniverse market prices.")

                return "Price preload completed!"



# Generated by Django 4.0.6 on 2022-09-09 14:01

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("buybackprogram", "0006_program_bonds_npc_price_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="program",
            name="expiration",
            field=models.CharField(
                choices=[
                    ("1 Day", "1 Day"),
                    ("3 Days", "3 Days"),
                    ("1 Week", "1 Week"),
                    ("2 Weeks", "2 Weeks"),
                    ("4 Weeks", "4 Weeks"),
                ],
                default="2 Weeks",
                help_text="Expiration time the contracts should bet set to.",
                max_length=7,
            ),
        ),
    ]



from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import F, Q, Value
from django.db.models.functions import Concat
from django.http import JsonResponse
from django.shortcuts import redirect, render
from django.utils.html import format_html
from eveuniverse.models import EveMarketGroup, EveSolarSystem, EveType

from allianceauth.services.hooks import get_extension_logger

from buybackprogram.forms import UserSettingsForm
from buybackprogram.models import Program, UserSettings
from buybackprogram.utils import messages_plus

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def index(request):
    user_groups = request.user.groups.all()

    logger.debug("User %s has groups: %s" % (request.user, user_groups))

    user_state = [request.user.profile.state]

    logger.debug("User %s state is : %s" % (request.user, user_state))

    try:
        user_settings = UserSettings.objects.get(user=request.user)
    except UserSettings.DoesNotExist:
        # create the default settings in the DB for the current user
        user_settings = UserSettings()
        user_settings.user = request.user
        user_settings.save()

        # get the user settings again
        user_settings = UserSettings.objects.get(user=request.user)

    program = (
        Program.objects.filter(
            Q(restricted_to_group__in=request.user.groups.all())
            | Q(restricted_to_group__isnull=True)
            | Q(owner__user=request.user)
        )
        .filter(
            Q(restricted_to_state=request.user.profile.state)
            | Q(restricted_to_state__isnull=True)
            | Q(owner__user=request.user)
        )
        .distinct()
    )

    context = {"programs": program}

    return render(request, "buybackprogram/index.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def faq(request):
    return render(request, "buybackprogram/faq.html")


@login_required
@permission_required("buybackprogram.manage_programs")
def item_autocomplete(request):
    items = EveType.objects.filter(published=True).exclude(
        eve_group__eve_category__id=9
    )

    q = request.GET.get("q", None)

    if q is not None:
        items = items.filter(name__icontains=q)

    items = items.annotate(
        value=F("id"),
        text=F("name"),
    ).values("value", "text")

    return JsonResponse(list(items), safe=False)


@login_required
@permission_required("buybackprogram.manage_programs")
def solarsystem_autocomplete(request):
    items = EveSolarSystem.objects.all()

    q = request.GET.get("q", None)

    if q is not None:
        items = items.filter(name__icontains=q)

    items = items.annotate(
        value=F("id"),
        text=F("name"),
    ).values("value", "text")

    return JsonResponse(list(items), safe=False)


@login_required
@permission_required("buybackprogram.manage_programs")
def marketgroup_autocomplete(request):
    items = EveMarketGroup.objects.all()

    q = request.GET.get("q", None)

    if q is not None:
        items = items.prefetch_related("parent_market_group").filter(name__icontains=q)

    items = items.annotate(
        value=F("id"),
        text=Concat(
            F("parent_market_group__parent_market_group__name"),
            Value(" -> "),
            F("parent_market_group__name"),
            Value(" -> "),
            F("name"),
        ),
    ).values("value", "text")

    print(items)

    return JsonResponse(list(items), safe=False)


@login_required
@permission_required("buybackprogram.basic_access")
def user_settings_edit(request):
    try:
        user_settings = UserSettings.objects.get(user=request.user)
    except UserSettings.DoesNotExist:
        # create the default settings in the DB for the current user
        user_settings = UserSettings()
        user_settings.user = request.user
        user_settings.save()

        # get the user settings again
        user_settings = UserSettings.objects.get(user=request.user)

    if request.method != "POST":
        user_settings_form = UserSettingsForm(instance=user_settings)
    else:
        user_settings_form = UserSettingsForm(request.POST, instance=user_settings)

        # check whether it's valid:
        if user_settings_form.is_valid():
            # user_settings.user = request.user
            user_settings.disable_notifications = user_settings_form.cleaned_data[
                "disable_notifications"
            ]
            user_settings.save()

            if user_settings.disable_notifications:
                messages_plus.success(
                    request,
                    format_html(
                        "Discord notifications <strong>disabled</strong>",
                    ),
                )
            else:
                messages_plus.success(
                    request,
                    format_html(
                        "Discord notifications <strong>enabled</strong>",
                    ),
                )

            return redirect("buybackprogram:index")

    context = {
        "form": user_settings_form,
    }

    return render(request, "buybackprogram/user_settings.html", context)



import json
from datetime import datetime

from django.contrib.auth.decorators import login_required, permission_required
from django.db.models import Q
from django.shortcuts import render
from django.utils import timezone
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.services.hooks import get_extension_logger

from buybackprogram.notes import (
    note_missing_from_contract,
    note_missing_from_tracking,
    note_quantity_missing_from_contract,
    note_quantity_missing_from_tracking,
)

from ..models import (
    Contract,
    ContractItem,
    ContractNotification,
    Tracking,
    TrackingItem,
)

from buybackprogram.app_settings import (
    BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS,
)

logger = get_extension_logger(__name__)


@login_required
@permission_required("buybackprogram.basic_access")
def my_stats(request):
    # List for valid contracts to be displayed
    valid_contracts = []

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
    }

    # Request user owned characters
    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(contract__issuer_id__in=characters)
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        # Add contract to the valid contract list
        valid_contracts.append(tracking)

    context = {
        "contracts": valid_contracts,
        "values": values,
        "mine": True,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def leaderboard(request, program_pk):
    # Tracker values
    monthstats = {
        "users": {},  # monthly stats per user
        "userinfo": {},  # profile information per user
        "months": None,  # all months
    }
    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(program_id=program_pk)
        .filter(contract__isnull=False)
        .prefetch_related("contract")
    )

    # Loop all tracking objects
    for tracking in tracking_numbers:
        # For finished contracts, gather more data
        if tracking.contract.status == "finished":
            month = datetime.strftime(tracking.contract.date_issued, "%Y-%m:%B %Y")
            if month not in monthstats["users"]:
                monthstats["users"][month] = {}  # User data per month

            # Collect ISK data per user
            user = tracking.contract.issuer_id
            if user not in monthstats["users"][month]:
                monthstats["users"][month][user] = [
                    0,  # contract total
                    0,  # donation total
                ]
            monthstats["users"][month][user][0] += tracking.contract.price
            monthstats["users"][month][user][1] += tracking.donation

            if user not in monthstats["userinfo"]:
                monthstats["userinfo"][user] = {
                    "name": EveEntity.objects.resolve_name(user),
                    "pic": f"https://images.evetech.net/characters/{user}/portrait?size=32",
                }

    monthstats["months"] = sorted(list(monthstats["users"].keys()))
    context = {
        "stats": json.dumps(monthstats),
    }

    return render(request, "buybackprogram/leaderboards.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_performance(request, program_pk):
    firstbench = datetime.now()
    lastbench = datetime.now()
    # Tracker values
    monthstats = {
        "status": {},
        "overall": {"all": {}},
        "items": {},
        "categories": {},
        "donations": {"all": {}},
    }
    allmonths = set()

    # Category to Item mapping
    category2items = {}

    # Export data
    dumpdata = [
        [
            "Contract ID",
            "Date Issued",
            "Date Finished",
            "User ID",
            "Total ISK",
            "Object Cateogry",
            "Object ID",
            "Object Name",
            "Object Quant",
            "Object ISK",
        ]
    ]

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(program_id=program_pk)
        .filter(contract__isnull=False)
        .prefetch_related("contract")
    )

    logger.debug(
        "Performance bench: %.2f" % (datetime.now() - lastbench).total_seconds()
    )
    lastbench = datetime.now()
    # Loop all tracking objects
    for tracking in tracking_numbers:
        month = datetime.strftime(tracking.contract.date_issued, "%Y-%m") + "-15"
        allmonths.add(month)

        if month not in monthstats["status"]:
            monthstats["status"][
                month
            ] = {}  # status of all contracts issued during a given month

        # Gather stats on all contracts' statuses
        if tracking.contract.status not in monthstats["status"][month]:
            monthstats["status"][month][tracking.contract.status] = 0
        monthstats["status"][month][tracking.contract.status] += 1

        # For finished contracts, gather more data
        if tracking.contract.status == "finished":
            # "overall": {"all": {"isk": {}, "q": {}, }
            if month not in monthstats["overall"]["all"]:
                monthstats["overall"]["all"][month] = [0, 0]  # isk, q
                monthstats["donations"]["all"][month] = [0, 0]  # isk, q

            monthstats["overall"]["all"][month][0] += tracking.contract.price
            monthstats["overall"]["all"][month][1] += 1

            monthstats["donations"]["all"][month][0] += tracking.donation
            if tracking.donation > 0:
                monthstats["donations"]["all"][month][1] += 1

            logger.debug(
                "Performance loop bench: %.2f"
                % (datetime.now() - lastbench).total_seconds()
            )
            lastbench = datetime.now()
            # Collect ISK data per items
            tracking_items = TrackingItem.objects.filter(tracking=tracking)
            for item in tracking_items:
                if item.eve_type.name not in monthstats["items"]:
                    monthstats["items"][item.eve_type.name] = {}
                if month not in monthstats["items"][item.eve_type.name]:
                    monthstats["items"][item.eve_type.name][month] = [0, 0]
                monthstats["items"][item.eve_type.name][month][0] += item.buy_value
                monthstats["items"][item.eve_type.name][month][1] += item.quantity
                # cache[tracking.contract.issuer_id] = EveEntity.objects.resolve_name(tracking.contract.issuer_id)

                # Collect item category data
                catid = item.eve_type.eve_group.name
                if catid not in category2items:
                    category2items[catid] = set()
                category2items[catid].add(item.eve_type.name)

                if catid not in monthstats["categories"]:
                    monthstats["categories"][catid] = {}
                if month not in monthstats["categories"][catid]:
                    monthstats["categories"][catid][month] = [0, 0]
                monthstats["categories"][catid][month][0] += item.buy_value
                monthstats["categories"][catid][month][1] += item.quantity

                # Collect data for export
                dumpdata.append(
                    [
                        tracking.id,
                        datetime.strftime(
                            tracking.contract.date_issued, "%Y-%m-%d %H:%M:%S"
                        ),
                        datetime.strftime(
                            tracking.contract.date_completed, "%Y-%m-%d %H:%M:%S"
                        ),
                        tracking.contract.issuer_id,
                        tracking.contract.price,
                        item.eve_type.eve_group.name,
                        item.eve_type.id,
                        item.eve_type.name,
                        item.quantity,
                        item.buy_value,
                    ]
                )

    logger.debug(
        "Performance exit loop bench: %.2f"
        % (datetime.now() - lastbench).total_seconds()
    )
    lastbench = datetime.now()
    # Reformat data so that it is easier to use billboard.js
    allmonths = sorted(list(allmonths))

    scaling = {}
    hscaling = {}
    for strata in ("overall", "items", "categories", "donations"):
        scaling[strata] = []
        for yi in monthstats[strata].keys():
            scaling[strata] += [
                monthstats[strata][yi][x][0] for x in monthstats[strata][yi].keys()
            ]
        if len(scaling[strata]) == 0:
            scaling[strata] = 1
        else:
            scaling[strata] = sum(scaling[strata]) / len(scaling[strata])
        for s, h in ((1e9, "Billions"), (1e6, "Millions")):
            if scaling[strata] > s:
                scaling[strata] = s
                hscaling[strata] = h
                break
        if scaling[strata] < 1e6:
            scaling[strata] = 1
            hscaling[strata] = ""

    # Always set donations to be the same scale as overall since they are plotted together.
    scaling["donations"] = scaling["overall"]
    hscaling["donations"] = hscaling["overall"]

    for strata in ("overall", "items", "categories", "donations"):
        for yi in monthstats[strata].keys():
            y = [[yi], [yi]]
            # Overall ISK
            if strata == "overall":
                y = [["Bought"], ["Contract count"]]
            # Donation ISK
            if strata == "donations":
                y = [["Donations"], ["Donation count"]]

            for m in allmonths:
                if m not in (monthstats[strata][yi]):
                    monthstats[strata][yi][m] = [0, 0]
                y[0].append(round(monthstats[strata][yi][m][0] / scaling[strata], 3))
                y[1].append(monthstats[strata][yi][m][1])
            monthstats[strata][yi] = y
    monthstats["x"] = ["x"] + allmonths

    for cat in category2items:
        category2items[cat] = list(category2items[cat])

    # Break down of categories by last three months:
    lastthree = {}
    for yi in monthstats["categories"].keys():
        calc = monthstats["categories"][yi][0].copy()
        calc.pop(0)
        lastthree[yi] = sum(calc[-3:])

    if len(lastthree.keys()) > 10:
        th = lastthree[sorted(lastthree.keys(), key=lambda x: -lastthree[x])[10]]
        lastthree["Other"] = 0
        for k in list(lastthree.keys()):
            if k != "Other" and lastthree[k] <= th:
                lastthree["Other"] += lastthree[k]
                del lastthree[k]

        monthstats["categories"]["Other"] = None
        for k in list(monthstats["categories"].keys()):
            if k not in lastthree:
                if monthstats["categories"]["Other"] is None:
                    monthstats["categories"]["Other"] = monthstats["categories"][k]
                else:
                    monthstats["categories"]["Other"][0][1:] = [
                        sum(x)
                        for x in zip(
                            monthstats["categories"]["Other"][0][1:],
                            monthstats["categories"][k][0][1:],
                        )
                    ]
                    monthstats["categories"]["Other"][1][1:] = [
                        sum(x)
                        for x in zip(
                            monthstats["categories"]["Other"][1][1:],
                            monthstats["categories"][k][1][1:],
                        )
                    ]
                del monthstats["categories"][k]
        monthstats["categories"]["Other"][0][0] = "Other"
        monthstats["categories"]["Other"][1][0] = "Other"

    lastthree = list(lastthree.items())

    # Sanitize CSV data
    for i in range(len(dumpdata)):
        for j in range(len(dumpdata[i])):
            dumpdata[i][j] = str(dumpdata[i][j]).replace(",", " ")

    context = {
        "stats": json.dumps(monthstats),
        "lastthree": json.dumps(lastthree),
        "categories": json.dumps(category2items),
        "export": json.dumps(dumpdata),
        "hscaling": json.dumps(hscaling),
    }
    logger.debug(
        "Performance finished: %.2f" % (datetime.now() - lastbench).total_seconds()
    )
    logger.debug(
        "Performance total: %.2f" % (datetime.now() - firstbench).total_seconds()
    )

    return render(request, "buybackprogram/performance.html", context)


@login_required
@permission_required("buybackprogram.manage_programs")
def program_stats(request):
    # List for valid contracts to be displayed
    valid_contracts = []
    untracked_contracts = False

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
        "untracked_count": 0,
    }

    # Request user owned characters
    characters = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__character_id", flat=True
    )

    # Request user owned corporations
    corporations = CharacterOwnership.objects.filter(user=request.user).values_list(
        "character__corporation_id", flat=True
    )

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(
            Q(contract__assignee_id__in=characters)
            | Q(contract__assignee_id__in=corporations)
        )
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        # Add contract to the valid contract list
        valid_contracts.append(tracking)

    if BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS:
        # Get pending contracts that have no tracking assigned to them
        untracked_contracts = Contract.objects.filter(
            Q(assignee_id__in=characters) | Q(assignee_id__in=corporations)
        ).filter(no_tracking=True, status="outstanding")

        logger.debug("Got %s untracked contracts" % len(untracked_contracts))

        for contract in untracked_contracts:
            values["untracked_count"] += 1

            # Get notes for this contract
            contract.notes = ContractNotification.objects.filter(contract=contract)

            # Get the name for the issuer
            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            # Get the name for the assignee
            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

    context = {
        "contracts": valid_contracts,
        "untracked_contracts": untracked_contracts,
        "values": values,
        "mine": True,
        "BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS": BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.see_all_statics")
def program_stats_all(request):
    # List for valid contracts to be displayed
    valid_contracts = []

    untracked_contracts = False

    # Tracker values
    values = {
        "outstanding": 0,
        "finished": 0,
        "outstanding_count": 0,
        "finished_count": 0,
        "untracked_count": 0,
    }

    # Get all tracking objects that have a linked contract to them for the user
    tracking_numbers = (
        Tracking.objects.filter(contract__isnull=False)
        .filter(contract__date_expired__gte=timezone.now())
        .prefetch_related("contract")
    )

    # Loop tracking objects to see if we have any contracts
    for tracking in tracking_numbers:
        # Get notes for this contract
        tracking.contract.notes = ContractNotification.objects.filter(
            contract=tracking.contract
        )

        # Walk the tracker values for contracts
        if tracking.contract.status == "outstanding":
            values["outstanding"] += tracking.contract.price
            values["outstanding_count"] += 1
        if tracking.contract.status == "finished":
            values["finished"] += tracking.contract.price
            values["finished_count"] += 1

        # Get the name for the issuer
        tracking.contract.issuer_name = EveEntity.objects.resolve_name(
            tracking.contract.issuer_id
        )

        # Get the name for the assignee
        tracking.contract.assignee_name = EveEntity.objects.resolve_name(
            tracking.contract.assignee_id
        )

        valid_contracts.append(tracking)

    if BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS:
        # Get pending contracts that have no tracking assigned to them
        untracked_contracts = Contract.objects.filter(
            no_tracking=True, status="outstanding"
        )

        logger.debug("Got %s untracked contracts" % len(untracked_contracts))

        for contract in untracked_contracts:
            values["untracked_count"] += 1

            # Get notes for this contract
            contract.notes = ContractNotification.objects.filter(contract=contract)

            # Get the name for the issuer
            contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

            # Get the name for the assignee
            contract.assignee_name = EveEntity.objects.resolve_name(
                contract.assignee_id
            )

    context = {
        "contracts": valid_contracts,
        "untracked_contracts": untracked_contracts,
        "values": values,
        "mine": True,
        "BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS": BUYBACKPROGRAM_TRACK_PREFILL_CONTRACTS,
    }

    return render(request, "buybackprogram/stats.html", context)


@login_required
@permission_required("buybackprogram.basic_access")
def contract_details(request, contract_title):
    contract = Contract.objects.get(title__contains=contract_title)

    # Get notes for this contract
    notes = ContractNotification.objects.filter(contract=contract)

    # Get items for this contract
    contract_items = ContractItem.objects.filter(contract=contract)

    # Get tracking object for this contract
    tracking = Tracking.objects.get(
        tracking_number=contract_title,
    )

    # Get tracked items
    tracking_items = TrackingItem.objects.filter(tracking=tracking)

    # Find the difference in the created contract and original calculation
    for tracking_item in tracking_items:
        tracking_notes = []

        item_match = False
        quantity_match = False

        for contract_item in contract_items:
            if contract_item.eve_type == tracking_item.eve_type:
                item_match = True

                if contract_item.quantity == tracking_item.quantity:
                    quantity_match = True
                    break

        tracking_item.item_match = item_match

        if not item_match:
            tracking_notes.append(note_missing_from_contract(tracking_item.eve_type))

        if item_match and not quantity_match:
            tracking_notes.append(
                note_quantity_missing_from_contract(tracking_item.eve_type)
            )

        tracking_item.notes = tracking_notes

    for contract_item in contract_items:
        contract_notes = []

        item_match = False
        quantity_match = False

        for tracking_item in tracking_items:
            if contract_item.eve_type == tracking_item.eve_type:
                item_match = True

                if contract_item.quantity == tracking_item.quantity:
                    quantity_match = True
                    break

        contract_item.item_match = item_match

        if not item_match:
            contract_notes.append(note_missing_from_tracking(contract_item.eve_type))

        if item_match and not quantity_match:
            contract_notes.append(
                note_quantity_missing_from_tracking(contract_item.eve_type)
            )

        contract_item.notes = contract_notes

    # Get the name for the issuer
    contract.issuer_name = EveEntity.objects.resolve_name(contract.issuer_id)

    # Get the name for the assignee
    contract.assignee_name = EveEntity.objects.resolve_name(contract.assignee_id)

    # Sort lists by reverse quantity order
    contract_items = sorted(contract_items, key=lambda x: -x.quantity)
    tracking_items = sorted(tracking_items, key=lambda x: -x.quantity)

    context = {
        "notes": notes,
        "contract": contract,
        "contract_items": contract_items,
        "tracking": tracking,
        "tracking_items": tracking_items,
    }

    return render(request, "buybackprogram/contract_details.html", context)
